/*
  Copyright (C) TomasVlad 2012

  Permission to use, copy, modify, distribute, and sell this
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in
  all copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaim all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

/*****************************************************
Chip type               : ATmega8
Program type            : Application
AVR Core Clock frequency: 10,000000 MHz
Memory model            : Small
External RAM size       : 0
Data Stack size         : 256
*****************************************************/

#include <mega8.h>
#include <delay.h>

#define ulong unsigned long
#define uint unsigned int

#define LCD_CMD  0
#define LCD_DATA 1

#define LCD_EN PORTB.0
#define LCD_RS PORTB.1
#define LCD_D4 PORTB.2
#define LCD_D5 PORTB.5
#define LCD_D6 PORTB.3
#define LCD_D7 PORTB.4

#define BTN_NULL    PIND.7
#define BTN_MODE    PIND.2
#define BTN_SET     PIND.6

#define RMODE  PORTD.1
#define CMODE 1
#define LMODE 0

#define CAL     PORTC.5
#define DIR_CAL DDRC.5
#define CK      100.0F   // input port cap == 10pF

#define RE     PORTC.3
#define DIR_RE DDRC.3

#define UEL  2  // ADC2
#define UESR 4  // ADC4 

#define DIR_IN  0
#define DIR_OUT 1

#define OMEGA 7     // Ohm

void msr_Uesr(void);
void msr_ElCm(void);

#pragma regalloc+
volatile uint Uesr_cnt;
volatile char cnt_TMR1_OVR;
volatile char cnt_TMR0_OVR;
volatile uint cnt_TMR1;
char Delay_3T @5;
uint Uesr @6;
#pragma regalloc-
// table
uint L0;
uint C0;
uint Kesr0;
uint Uesr0;
uint Re0;
ulong Cp0;
uint Ccal0;
uint KUesr;

char eeprom *EE_str_off;
char lcd_LineN;
char lcd_Nybble;
char str_pos_now;
float F1pwr2;
float F0pwr2;
float FXpwr2;
float k;
ulong Cx;
uint Fx;
uint dFx;

uint ElCm_time;
uint Ucm0;
uint Ucm1;
uint UcmTop;
uint Utmp;
uint dUcm;

char btn_Nul;
char btn_Set;
char btn_Mod;

char dUc_200mS;
char dUc_100mS;
char value_exponent;
char tmp_idx = 0;
char str_pos_end;
char value_units;
uint Kesr;
char Uesr_OVR;
char tgCap;
ulong lcd_data;
char str[12];

// flags_mode
bit fmInitTmr;
bit fmRdyOk;
bit fmElCx;
bit fmLCx;
bit fmC33n;
bit fmFreqStab;
bit fmIc;
bit fmRx;

// flags_stat
bit fsDelay_1T;
bit fsDelay_2T;
bit fsFreqChange;
bit fsMinus;
bit fsSet;
bit fsDig3;
bit fsCgAdr0;
bit fsCgAdr1;

#pragma keep+
eeprom char conf_lcd[] =    { 0x28,  // 4-bit interface, dual line, 5x7 dots
                            0x8,    // display off
                            0xC,    // display on, cursor off
                            0x1,    // clear display
                            0x6,    // display shift off, inc cursor move dir
                            0xFF };
eeprom char a_Cx[] =        {' ',0x5,' ','C','x',0xFF };
eeprom char a_Cp[] =        {' ',0x5,' ','C','p',0xFF };
eeprom char a_vC0[] =       {' ',0x5,' ','C','0',0xFF };
eeprom char a_Lx[] =        {' ',0x6,' ','L','x',0xFF };
eeprom char a_vL0[] =       {' ',0x6,' ','L','0',0xFF };
eeprom char a_ElCm[] =      {0x3,' ','C',' ','m',0xFF };
eeprom char a_rESR[] =      {0x4,' ','E','S','R',0xFF };
eeprom char a_Ccal[] =      {0x3,'C','c','a','l',0xFF };
eeprom char a_KUesr[] =     {'K','U','e','s','r',0xFF };
eeprom char a_F0[] =        {' ','F','0',' ',' ',0xFF };
eeprom char a_Fcal[] =      {' ','F','C','a','l',0xFF };
eeprom char a_ElCsz[] =     {0x3,'C',' ','s','z',0xFF };
eeprom char a_rRx[] =       {' ',0x4,' ','R','x',0xFF };
eeprom char a_lcLx[] =      {0x5,0x6,' ','L','x',0xFF };
eeprom char a_Uesr0[] =     {'U','e','s','r','0',0xFF };
eeprom char a_Uesr[] =      {' ','U','e','s','r',0xFF };
eeprom char a_Fesr[] =      {' ','F','e','s','r',0xFF };
eeprom char a_icEsr[] =     {0x2,'-','E','S','R',0xFF };
eeprom char a_Re[] =        {' ',0x4,' ','R','e',0xFF };
eeprom char a_Ro[] =        {' ',0x4,' ','R','o'};
eeprom char a_0xFF[] =      {0xFF };
eeprom char a_C0[] =        {' ',' ',' ','C','-','-','-','>','0',' ',' ',' ',' ',' ',' ',' ',0xFF };
eeprom char a_L0[] =        {' ',' ',' ','L','-','-','-','>','0',' ',' ',' ',' ',' ',' ',' ',0xFF };
eeprom char a_Err[] =       {' ',' ',' ','>','>','!','E','r','r','o','r','!','<','<',' ',' ',0xFF };
eeprom char a_ToSet[] =     {'=','=','=',' ','S','e','t',' ','M','o','d','e',' ','=','=','=',0xFF };
eeprom char a_ToNul[] =     {'=','=','=',' ','N','u','l',' ','M','o','d','e',' ','=','=','=',0xFF };
eeprom char a_ban0[] =      {'L','C','M','3','-','M','E','G','A','8',' ','C','L','O','N','E',0xFF };
eeprom char a_ban1[] =      {'v','.','1','.','5','d',' ','(','c',')','T','o','m','a','s','V',0xFF };
eeprom char zgn_lcd[] =     { 0x0, 0x0, 0x3, 0x4, 0x2, 0x5, 0x9, 0x6,               // "delta"
                              0x18, 0x8, 0x18, 0xB, 0x1A, 0x3, 0x1, 0x3,            // "3/5"
                              0x0, 0x0, 0x13, 0x14, 0x14, 0x14, 0x13, 0x0,          // "ic"
                              0x4, 0x1F, 0x11, 0x1F, 0x0, 0x1F, 0x4, 0x0,           // "electrolit C"
                              0x4, 0xE, 0xA, 0xA, 0xA, 0xE, 0x4, 0x0,               // "Resistor"
                              0x4, 0x4, 0x1F, 0x0, 0x1F, 0x4, 0x4, 0x0,             // "Capicator"
                              0x8, 0x1D, 0x1D, 0x1D, 0x1D, 0x1D, 0x8, 0x0,          // "Inductor"
                              0x0, 0xE, 0x11, 0x11, 0x11, 0xA, 0x1B, 0x0, 0xFF };   // "Ohm"
eeprom char tbl_const[] =   { 0x00, 0x26,               // L0=9728 [pF*10]
                              0xCA, 0x17,               // C0=5322 [pF*10]
                              0xDD, 0x37,               // Kesr=
                              0x51, 0x0C,               // Uesr0=3153 [mV*10]
                              0xB9, 0x00,               // Re0=185 [Ohm]
                              0x2C, 0x06, 0x05, 0x00,   // Cp0=329260 [pF*10]
                              0x60, 0x23,               // Ccal0=9056 [pF*10]
                              0x62, 0x01 };             // KUesr=354
                                                    
#pragma keep-

eeprom char prefix[] = { 'p','n','u','m',' ','k',0x0 };

// ext T0 pin Falling Edge count (8-bit)
interrupt [TIM0_OVF] void timer0_ovf_isr(void)
{
	if(fmInitTmr) {
		TCNT1 = 0;      // TMR1H=0;TMR1L=0;
		cnt_TMR1_OVR=0;
		TCNT0 = 0xC0;   // = -64
		fmInitTmr = 0;
	} else {
		cnt_TMR1 = TCNT1;
		cnt_TMR0_OVR++; // pulse_count++
		
		if(cnt_TMR1_OVR > 7) {  // cnt = 8
			fmRdyOk=1;
			if((cnt_TMR1>>8) > 0) TIMSK &= ~(1<<TOIE1); // disable TMR1IE
		}
		
	}
}

// service buttons
interrupt [EXT_INT0] void ext_int0_isr(void)
{
    GICR &= ~(1<<INT0);     // disable INT0
    TCNT2 = 0;              // clear TCNT2
    SFIOR |= (1<<PSR2);     // reset Timer2 prescaler
    TIMSK |= (1<<TOIE2);    // enable Timer2_INT
}

// time count dT=100nSec (16-bit)
interrupt [TIM1_OVF] void timer1_ovf_isr(void) { cnt_TMR1_OVR++; }

// dT=26mS
interrupt [TIM2_OVF] void timer2_ovf_isr(void)
{ 
    TIMSK &= ~(1<<TOIE2);    // disable Timer2_INT

    if(!BTN_MODE) {
        if(fsSet) {
            if(!BTN_NULL) btn_Nul ^= 1;
            else if(!BTN_SET) btn_Set = 1;
            else btn_Mod ^= 1;
            if(btn_Mod && btn_Nul) { btn_Mod = 0; btn_Nul = 0; }
        } else {
            if(!BTN_NULL) btn_Nul = 1;
            else if(!BTN_SET) btn_Set = 1;
            else btn_Mod = 1;
        }
    }

    GIFR |= (1<<INTF0);     // clear INTF0
    GICR |= (1<<INT0);      // enable INT0
}

// measure Uesr
interrupt [EXT_INT1] void ext_int1_isr(void)
{
	char i;
    GICR &= ~(1<<INT1);     // disable INT1
    // ??? test ???
    #asm {
        LDS  R30,_Uesr_cnt  // 2T
        ANDI R30,LOW(0x2)   // 1T
        BREQ PC+3           // 1/2T
        NOP                 // 1T
        NOP                 // 1T
    }
    #endasm
    
//    #asm("nop");
//    #asm("nop");

    #asm {
        MOV  R16,_Delay_3T
loop:
        SUBI  R16,1 // 1T
        BRNE loop   // 1/2T
        
        BST  R3,0   // 1T
        BRTC d2T    // 1/2T
        NOP
        NOP
d2T:
        BST  R3,1
        BRTC ddd
        NOP
        NOP
        NOP
ddd:
        NOP        
    }
    #endasm

    ADCSRA |= (1<<ADSC);    // (2T)ADC Start Conversion
	
    #asm("nop");
    #asm("nop");
    
    while((ADCSRA & (1<<ADIF))==0);     // wait ADSC==0
    ADCSRA |= (1<<ADIF);    // clear ADIF
    
    #asm
        .EQU ADCL=0x4
        .EQU ADCH=0x5
    #endasm
    
	if(!(Uesr_cnt & 1)) {   // Uesr += ADCW
		#asm {
            IN R30, ADCL
            IN R31, ADCH
            ADD R6, R30
            ADC R7, R31
        }
        #endasm

        MCUCR |= (1<<ISC11) | (1<<ISC10);   // The rising edge of INT1 generates an interrupt request
	} else {                // Uesr -= ADCW
		#asm {
            IN R30, ADCL
            IN R31, ADCH
            SUB R6, R30
            SBC R7, R31
        }
        #endasm

		MCUCR &= ~(1<<ISC10);   // The falling edge of INT1 generates an interrupt request.
	}
     
	Uesr_cnt++;     // cnt_int0++
	
	if((Uesr & (1<<13)) && (Uesr & (1<<12))) {  // Uesr==12288
	    Uesr_OVR++;
	    Uesr &= ~(1<<13);   // Uesr = 12288 - 8192 = 4096
    }

    GIFR |= (1<<INTF1);     // clear INTF1
    if(Uesr_cnt < 3200) GICR |= (1<<INT1);  // enable INT1 
    GIFR |= (1<<INTF1);     // clear INTF1

    i=0;
    while(i<64) {
        if(GIFR & (1<<INTF1)) break;
        i++;
    }

    GIFR |= (1<<INTF1);     // clear INTF1
}

void lcd_wrNybble(void)
{
		LCD_D4 = (lcd_Nybble & 1) ? 1 : 0;
		LCD_D5 = (lcd_Nybble & 2) ? 1 : 0;
		LCD_D6 = (lcd_Nybble & 4) ? 1 : 0;
		LCD_D7 = (lcd_Nybble & 8) ? 1 : 0;
		delay_us(5);
		LCD_EN = 1;
		delay_us(5);
		LCD_EN = 0;
		delay_us(5);
}

void lcd_wrByte(char data, char flg)
{
	LCD_RS=0;
	delay_us(50);
	
	if(flg) LCD_RS=1;    // write data
	
	lcd_Nybble = data >> 4;
    lcd_wrNybble();
    lcd_Nybble = data;
    lcd_wrNybble(); 
    delay_us(50);
	if(!str_pos_now) LCD_RS=0;
}

void lcd_EEdata_to_DDRAM(char eeprom *EE_adr)
{
	char ii;
    char eeprom *adr;
	
	if(lcd_LineN != 0xFF) { 
		ii = 0x80;
		if(lcd_LineN) ii |= 0x40;   // lcd_line=1
		lcd_wrByte(ii, LCD_CMD);    // set display line
	}

	adr=EE_adr;
	while(1) {
		ii = *adr;
		if(ii==0xFF) break;    // end data
		if(EE_adr==conf_lcd) lcd_wrByte(ii, LCD_CMD);
		else lcd_wrByte(ii, LCD_DATA);
		adr++;
	}	

	if(adr == EE_adr) {     // LCD clear line
		for(ii=0; ii<16; ii++) lcd_wrByte(' ', LCD_DATA);	
	}
}

void dis_btn_int(void)
{
    #asm("cli");
    GICR &= ~(1<<INT0);     // disable INT0
    TIMSK &= ~(1<<TOIE2);   // disable Timer2_INT 
}

void ena_btn_int(void)
{
    GICR |= (1<<INT0);      // INT0 Enable
    #asm("sei");
}

void msr_Freq(void)
{
	float val;
    uint adc_dly;

	dis_btn_int();
    delay_ms(100);
    FXpwr2 = 0;
    Fx = 0;
    fmInitTmr = 1;
	fmRdyOk = 0;
	cnt_TMR0_OVR = 0xFF;
	cnt_TMR1_OVR = 0;
	TCNT0 = 0xFE;
	TCNT1 = 0;
	TIFR |= (1<< TOV0);     // T0IF = 0;
	// Enable interrupts
	TIMSK |= (1<<TOIE1);    // Timer1 Overflow Interrupt Enable
	TIMSK |= (1<<TOIE0);    // Timer0 Overflow Interrupt Enable bit
	#asm("sei");             // Global Interrupt Enable
	delay_ms(50);

	if(!fmInitTmr) {
		if(fmC33n)  // C33n==On
			while(!fmRdyOk && !(cnt_TMR1_OVR & 0x80));
		else        // C33n==Off
			while(!fmRdyOk && !(cnt_TMR1_OVR > 39));
	}

	#asm("cli");     // Global Interrupt Disable
	TIMSK &= ~(1<<TOIE0);    // disable Timer0 Overflow Interrupt
	TIMSK &= ~(1<<TOIE1);    // disable Timer1 Overflow Interrupt
	ena_btn_int();
    
	if(!fmRdyOk) return;

//    time = ((ulong)cnt_TMR1_OVR<<16) + cnt_TMR1 - DLY_TMR1;   // =t[0.2 uS] DLY_TMR1==54
//    time = (float)(((ulong)cnt_TMR1_OVR<<16) + cnt_TMR1);     // =t[0.2 uS]
//    tmp =6400000.0*cnt/time;
    
    val = 6400000.0*(float)(((ulong)cnt_TMR0_OVR<<2)+1);    // cnt_TMR_OVR == pulse_count*256 (cnt_TMR1_OVR > 7 --> stop)
    val /= (float)(((ulong)cnt_TMR1_OVR<<16) + cnt_TMR1);   // TMR1 == time_count(dT==100nS)
	Fx = val;                   // = F[kHz*10]
	FXpwr2 = val*val;           // = Fx^2[kHz^2*100]

	if(fmC33n) FXpwr2 -= 0.6;   // C33n==On
	
    Kesr = FXpwr2/55.0;         // = Fx^2/55.0
	
    val = (10000.0/(Fx*0.1))*1.75;
    adc_dly = (uint)(val - 69.0);

	Delay_3T = 0;
    while(adc_dly > 2) {
		adc_dly -= 3;
		Delay_3T++;
	}
	// ??? test ???
    Delay_3T--;
	
    fsDelay_1T = adc_dly & 1;
    fsDelay_2T = (adc_dly >> 1) & 1;
}

void lcd_show_value(void)
{
	char chr_pos, ii, i;
	uint tmp_data;

	chr_pos = 7;
    for(i=0;i<12;i++) str[i]=' ';
	str[7]='?';
    str[11]=0;
	
	if(fsDig3) str[2] = '=';
	else str[1] = '=';
	
	if((a_rESR == EE_str_off) && (fmRdyOk)) {
            str[0]=0;   // ==zgn_delta
            str[1]=1;   // ==zgn_graph
        }
	
	fsMinus = 0;
    if(lcd_data & (1<<31)) {
        fsMinus = 1;
        lcd_data = (uint)(0 - lcd_data);
    }
	
//	if(lcd_data < 3) lcd_data = 0;  // dead "0" from +-2
    
    if(lcd_LineN == 0) {
		tmp_data = lcd_data;
        if(fmElCx) ii = 0;
        else ii = 4;
start_shift:		
		if(tmp_data > 16) goto next_shift;
        if(ii > 6) goto end_shift;
next_shift:
			tmp_data >>= 1;
			ii++;
            goto start_shift;
end_shift:		
		ii -= 7;
		ii = (ii<<4)|(ii>>4);	// swap
		ii += (tmp_data & 0xFF);
		
        if(!fmRdyOk) ii=0xFF;
		tmp_idx += 2;
		tmp_idx -= ii;
		
		if(str_pos_now) {
			fsCgAdr0 ^= 1;
			if(!fsCgAdr0) fsCgAdr1 = 1;
			if(fsCgAdr0 && !fsCgAdr1) str_pos_now--;
		}
		
		if(tmp_idx >= 4) {
			tmp_idx = str_pos_end - str_pos_now;
			tmp_idx >>= 2;
			tmp_idx &= 0x3F;
			while(tmp_idx) {
				tmp_idx >>= 1;
				if(str_pos_end != 0xFF)	str_pos_end++;	
			}
			if(!str_pos_now) str_pos_now = str_pos_end;
		}
		
		tmp_idx = ii;
	}   // end lcd_LineN == 0		
		
	if(!fmRdyOk) goto copy_str;     // fmRdyOk==0

    // fmRdyOk == 1	
    while(lcd_data > 9743) {
        lcd_data /= 10;
        value_exponent++;
    }
		
    ii = 0;
	while(value_exponent > 2) {
        value_exponent -= 3;
        ii++;
    }
	
    // convert to dec	
	value_exponent += 4;
	if(fsDig3) value_exponent++;   // str[2] = '='
	if(value_exponent == 7) value_exponent = 0;
	
	// convert data to char's
	while(1) {
		if(chr_pos == value_exponent) {    // init: pos==7
			str[chr_pos] = '.';
            chr_pos--;
			value_exponent = 0;
		}
		
		tmp_data = 0;
		while(lcd_data > 9) {
			lcd_data -= 10;
			tmp_data++;	
		}
		
		str[chr_pos] = lcd_data + '0';
        chr_pos--;
		
        if(fsDig3) {    // str[2] = '='
            chr_pos = 7;
            fsDig3 = 0;                                                         
        }
		
        lcd_data = tmp_data;
		if(lcd_data) continue;
        if(!value_exponent) break;
	} 
	
	if(chr_pos == 7) chr_pos--;
	if(fsMinus) str[chr_pos] = '-';
	str[9] = prefix[ii];     // idx='p','n','u','m',' ','k'
	if(dUc_100mS) { str[0] = dUc_200mS; str[1]= dUc_100mS; }
	
copy_str:    
    str[10] = value_units;
	dUc_100mS = 0;	

    // show LCD display data
	lcd_EEdata_to_DDRAM(EE_str_off);
    for(i=0;i<11;i++) lcd_wrByte(str[i], LCD_DATA);
}

void cal_L0_C0(void)
{
    float sav_F1pwr2;
    float tmpC;
    
    sav_F1pwr2 = F1pwr2;
    F1pwr2 = FXpwr2;
    
    DIR_CAL = DIR_OUT;  // Ccal = ON
    delay_ms(500);
    msr_Freq();         // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (fmC33n=1 -> Fx^2-0.6)
    DIR_CAL = DIR_IN;   // Ccal = OFF

    F0pwr2 = FXpwr2;
    
    // calc C0
//    Cx = (ulong)((float)Ccal0/(F1pwr2/F0pwr2 - 1.0));
    k = F1pwr2 / F0pwr2;
    Cx = (ulong)((((float)Ccal0 - k * CK)/(k - 1.0)) + CK);    
//    Cx = (ulong)((float)Ccal0/(kF - 1.0));    
    
    F0pwr2 = F1pwr2;
    
    if(RMODE == LMODE) F1pwr2 = sav_F1pwr2;
    
    // Lx=(1/(2*pi)^2)/(Cx*F0^2)

    tmpC = 2.533029591e+14F;    // C33n==On
    if(!fmC33n) tmpC *= 10.0;   // C33n==Off
    
    // calc L0
    lcd_data = (ulong)(tmpC / ((float)Cx * F0pwr2));
    
    if(RMODE == LMODE) {      // mode==LX
        L0 = lcd_data;  // L0=Lx
        L0--;
        C0 = Cx;
        C0++;
    }
}

uint msr_DC_ADC(void)
{
    char i;
    uint result;
    
    delay_us(20);
    result = 3;

    for(i=0;i<49;i++) {
        ADCSRA |= (1<<ADSC);                // ADC Start Conversion
	    while((ADCSRA & (1<<ADIF))==0);     // wait ADIF==1
        ADCSRA |= (1<<ADIF);                // clear ADIF
        result += ADCW;
    }
    return (result &= 0xFFF8);
}
/*
void check_btn(void)
{
    if(btn_Mod) RMODE ^= 1;
    btn_Mod = 0;
}
*/
void init_lcm(void)
{
	char *ptr;
    char i;

    ADMUX |= UEL;               // ADC.2(ElCx)
    DIR_RE = DIR_OUT; RE = 0;   // Re=GND
    RMODE = CMODE;
    fsSet = 0;
    
// ====== configure LCD ==========
    // wait 50ms or more after power-on
    delay_ms(50);
    // set 8-bit mode
    lcd_Nybble = 3;
    lcd_wrNybble(); // Hi-Nybble
    delay_ms(5);
    lcd_wrNybble(); // Lo-Nybble
    delay_ms(5);
    // set 4-bit mode
    lcd_wrNybble(); // Hi-Nybble
    delay_ms(5);
    lcd_Nybble = 2;
    lcd_wrNybble(); // Lo-Nybble
    delay_ms(5);
    // Configure LCD interface    
	lcd_EEdata_to_DDRAM(conf_lcd);
	delay_ms(10);
	lcd_wrByte(0x40,LCD_CMD);	// Cmd=0x40(Sets CGRAM address 0x0)
	lcd_LineN=0xFF;             // no set LCD_LINE!
    // load LCD graphic's to CGRAM
    lcd_EEdata_to_DDRAM(zgn_lcd); 
   	DIR_RE = DIR_IN;    // Re=OFF
    // show banner
    lcd_LineN=0;
	lcd_EEdata_to_DDRAM(a_ban0);
    lcd_LineN=1;
	lcd_EEdata_to_DDRAM(a_ban1);
	delay_ms(200);
    
    btn_Nul = 0;
    btn_Set = 0;
    btn_Mod = 0;
    RMODE = CMODE;
    
    lcd_LineN = 0;
	str_pos_end = 64;
	str_pos_now = 64;
    // Load constants from EEPROM
    ptr = (char*)&L0;
	for(i=0;i<18;i++) *(ptr+i) = tbl_const[i];
	delay_ms(500);
    // measure & show F0
    msr_Freq();
	EE_str_off = a_F0;
	value_units = ' ';
	value_exponent = 17;
	fsDig3 = 0;         // str[1] = '='
	dUc_100mS = 0;
	lcd_data = Fx;
    lcd_show_value();
    // measure & show Fcal    
    cal_L0_C0();
    lcd_data = Fx;
    lcd_LineN = 1;
    value_exponent = 17;
    EE_str_off = a_Fcal;
    lcd_show_value();
    delay_ms(2000);
}

void msr_Uesr(void)
{
	Uesr_cnt = 0;
	Uesr = 15;
	dUc_100mS = 0;
	Uesr_OVR = 0;
	fmRdyOk=0;
	
	if(Fx < 650) return;   // < 65khZ
	if(Fx > 950) return;   // > 95khZ

	dis_btn_int();
    
    ADMUX &= 0xF0;      // ADC.0
    ADMUX |= UESR;      // ADC.4(Uesr)
	delay_us(10);
    
    MCUCR &= ~(1<<ISC10);
    MCUCR |= (1<<ISC11);    // The falling edge of INT1 generates an interrupt request.
	GICR |= (1<<INT1);      // External Interrupt Request 1 Enable
 	
    #asm("sei");
    delay_ms(100);
	GICR &= ~(1<<INT1);     // External Interrupt Request 1 Disable
	#asm("cli");
	
	ADMUX &= 0xF0;      // ADC.0
    ADMUX |= UEL;       // ADC.2(ElCx)

    ena_btn_int();

	Uesr >>= 5;     // Uesr /= 32
    while(Uesr_OVR) { Uesr += 256; Uesr_OVR--; }
    
    //  ==Uesr[mV*10]
	if(Uesr > 30) fmRdyOk=1;        // Uesr > 3.0[mV] 
}

void msr_Cp(void)
{
    char F_C33_On, F_C33_Off;
    float sav_F1pwr2;
    
    sav_F1pwr2 = F1pwr2;  // save F1^2
    fmLCx = 0;

    // msr_Freq(C33_On)
    DIR_RE = DIR_OUT;   // Re=GND
    fmC33n = 1;         // C33n==On
    msr_Freq();         // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (C33n==On -> Fx^2-0.6)

    // check Freq(C33n=On)    
    F_C33_On = Fx & 0xFF;            // F_C33_On=Fx(C33_On);
    if(Fx >> 8) F_C33_On = 0xFF;     // Fx(ReOn) > 25.5[kHz]
    if(!F_C33_On) F_C33_On++;
    if(!fmRdyOk) F_C33_On = 0;

    // msr_Freq(C33_Off)
    DIR_RE = DIR_IN;    // Re=Off
    fmC33n = 0;         // C33n==Off
    msr_Freq();         // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (C33n==On -> Fx^2-0.6)

    // check Freq(C33_Off)    
    F_C33_Off = (Fx >> 3) & 0xFF;         // F_C33_Off=Fx(C33_Off)/8
    if((Fx >> 8) > 7) F_C33_Off = 0xFF;   // Fx(ReOff) > 179.2[kHz]
    if(!F_C33_Off) F_C33_Off++;
    if(!fmRdyOk) F_C33_Off = 0;
    
    if(!F_C33_On) goto cont;              // F_C33_On ==0
    if(F_C33_On > 59) goto cont;          // F_C33_On > 5.9Khz
    if(F_C33_On > 23)                     // F_C33_On > 2.3kHz
        if(F_C33_On > F_C33_Off) goto cont; // F_C33_On > F_C33_Off/8
    if(F_C33_On > 7) goto msr_ReOn;       // F_C33_On > 0.7kHz
    if(F_C33_Off > 16) goto cont;         // F_C33_Off > 12.8kHz

msr_ReOn:
    // msr_Freq(C33_On)
    DIR_RE = DIR_OUT;   // Re -> GND
    fmC33n = 1;         // C33n=On
    msr_Freq();         // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (C33n==On -> Fx^2-0.6)
    F_C33_Off = 0;
    fmRdyOk = 1;

//-----------------------------------------    
cont:
    F0pwr2 = FXpwr2;    // F0^2=Fx^2(CcalOff && ReOn)
    
    if(F_C33_Off) { fmRdyOk = 1; F_C33_On = 0; }  // F_C33_Off > 0
    if(F_C33_On > 7) goto msr_CcalOn;             // F_C33_On > 0.7kHz 
    if(F_C33_Off > 127) goto exit;                // F_C33_Off > 101.6kHz
    if(!F_C33_Off) goto exit;                     // F_C33_Off ==0

    // !!! measure parasitic Cx !!!    
msr_CcalOn:
    F1pwr2 = F0pwr2;  // F1^2=F^2(CcalOff && ReOn)

    // msr_Freq(CcalOn && ReOff)
    DIR_CAL = DIR_OUT;  // Ccal=On
    delay_ms(500);
    msr_Freq();         // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (C33n==On -> Fx^2-0.6)
    DIR_CAL = DIR_IN;   // Ccal=Off

    F0pwr2 = FXpwr2;    // F0^2=Fx^2(CcalOn && ReOff)
    // calc Cx    
//    Cx = (float)Ccal0 / (F1pwr2 / F0pwr2 - 1.0);
    k = F1pwr2 / F0pwr2;
    Cx = ((float)Ccal0 - k * CK) / (k - 1.0) + CK;
//    Cx = (float)Ccal0 / (kF - 1.0);

    fmLCx = 1;          // !!! show parasitic Cx !!!
    
exit:
    F1pwr2 = sav_F1pwr2;  // restore F1^2
}

void set_mode(void)
{

    fsSet = 1;
    btn_Set = 0;
    btn_Nul = 0;
    btn_Mod = 0;
    fmRdyOk = 1;
    lcd_LineN = 0;
    
    lcd_EEdata_to_DDRAM(a_ToSet);
    lcd_LineN = 1;

set_ccal:
        if(btn_Nul)     // BTN_PLUS
            if((Ccal0 >> 8) < 46) Ccal0++;
        if(btn_Mod)    // BTN_MINUS
            if((Ccal0 >> 8) > 27) Ccal0--;
        
        value_units = 'F';
        lcd_data = Ccal0;
        EE_str_off = a_Ccal;
        value_exponent = 2;     // [pF]
        str_pos_now = str_pos_end;
        lcd_show_value();
        delay_ms(100);
        if(btn_Set) goto set_re;
        goto set_ccal;

set_re:
        btn_Nul = 0;
        btn_Mod = 0;
        btn_Set = 0;
set_re_:        
        if(btn_Nul)     // BTN_PLUS
            if(Re0 < 250) Re0++;
        if(btn_Mod)    // BTN_MINUS
            if(Re0 > 150) Re0--;
        
        value_units = OMEGA;    // "Ohm"
        lcd_data = Re0;
        EE_str_off = a_Re;
        value_exponent = 15;
        fsDig3 = 0;             // str[1] = '='
        lcd_show_value();
        delay_ms(100);
        if(btn_Set) goto set_kuesr;
        goto set_re_;

set_kuesr:
        btn_Nul = 0;
        btn_Mod = 0;
        btn_Set = 0;
set_kuesr_:        
        if(btn_Nul)     // BTN_PLUS
            if(KUesr < 700) KUesr++;
        if(btn_Mod)    // BTN_MINUS
            if(KUesr > 100) KUesr--;
        
        value_units = ' ';
        lcd_data = KUesr;
        EE_str_off = a_KUesr;
        value_exponent = 14;
        str_pos_now = str_pos_end;
        lcd_show_value();
        delay_ms(100);
        if(btn_Set) goto end_set;
        goto set_kuesr_;
end_set:        
        fsSet = 0;
        btn_Nul = 0;
        btn_Mod = 0;
        btn_Set = 0;
        lcd_EEdata_to_DDRAM(a_0xFF);
        delay_ms(100);
}

void nul_mode(void)
{
    char i;
    char *ptr;
    
    btn_Nul = 0;
    
    lcd_LineN = 1;
    lcd_EEdata_to_DDRAM(a_0xFF);
    lcd_LineN = 0;
    lcd_EEdata_to_DDRAM(a_ToNul);
    delay_ms(1000);

    if(!fmRdyOk) {
        lcd_EEdata_to_DDRAM(a_Err);
        goto exit;
    }
    
    lcd_LineN = 0;
    str_pos_now = str_pos_end;

    delay_ms(500);
    
    msr_Freq();     // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (C33n==On -> Fx^2-0.6)
    
    if(Fx > 1050) {      // >105kHZ
        cal_L0_C0();
        fmElCx = 0;
        lcd_LineN = 0;
        lcd_EEdata_to_DDRAM(a_0xFF);
        if(RMODE == CMODE) {
            value_units = 'F';
            value_exponent = 2;
            EE_str_off = a_vC0;
            lcd_data = Cx;
            lcd_show_value();
            lcd_LineN = 1;
            lcd_EEdata_to_DDRAM(a_C0);
        } else {
            value_units = 'H';
            value_exponent = 7;
            EE_str_off = a_vL0;
            lcd_show_value();
            lcd_LineN = 1;
            lcd_EEdata_to_DDRAM(a_L0);
        }
    } else {
        Kesr0 = Kesr;
        msr_Uesr();
        Uesr0 = Uesr;
        msr_Cp();       // res=Cx; set fmLCx!!!
        Cp0 = Cx;
        
        //show Ro0
        lcd_data = L0;
        lcd_data *= L0;
        lcd_data >>= 16;
        lcd_data *= Kesr;
        lcd_data /= Uesr0;
        
        lcd_LineN = 0;
        value_units = OMEGA;    // "Ohm"
        value_exponent = 11;
        EE_str_off = a_Ro;
        lcd_show_value();

       // show Cp
        lcd_LineN = 1;
        value_units = 'F';
        value_exponent = 2;
        EE_str_off = a_Cp;
        lcd_data = Cx;
        lcd_show_value();
        delay_ms(2000);
        
        // show Uesr    
        lcd_LineN = 0;
        value_units = 'V';
        value_exponent = 11;
        EE_str_off = a_Uesr0;
        fsDig3 = 0;     // str[1] = '='
        lcd_data = Uesr0;
        lcd_show_value();
        
        lcd_LineN = 1;
        value_units = ' ';
        value_exponent = 17;
        EE_str_off = a_Fesr;
        lcd_data = Fx;
        lcd_show_value();    
    }
    // write to EEPROM
    ptr = (char*)&L0;
	for(i=0;i<18;i++) tbl_const[i] = *(ptr+i);
exit:    
    delay_ms(1000);
}

void msr_ElCm(void)
{
    uint Ux;
    
    ElCm_time = 0;
    lcd_data = 0;
    
    if(fmRx || fmIc) {
        delay_ms(500);
        return;
    }
    
    EE_str_off = a_ElCm;
    fmElCx = 0;
    
    DIR_RE = DIR_OUT; RE = 0;   // Re=GND
    delay_ms(500);              // discharge elCm 500mS    
    DIR_RE = DIR_IN;            // Re=Off
    Ucm0 = msr_DC_ADC();        // ADC.2
        
    if(Ucm0 & (1<<15)) goto discharge_Cm_5mS;   // Ucm0 > 3276.8[mV]
        
    Ucm1 = (Ucm0 + 10000) / 49;     // Ucm1 = Ucm0 + 1000.0[mV]
        
    // charge Cm to Ucm0 + 1000.0[mV]
    DIR_RE = DIR_OUT; RE = 1;   // Re=VCC
    
    while(1) {
        ElCm_time++;
        ADCSRA |= (1<<ADSC);                // ADC Start Conversion
	    while((ADCSRA & (1<<ADIF))==0);     // wait ADSC==0
        ADCSRA |= (1<<ADIF);
        if((ElCm_time & (1<<14)) || (ADCW >= Ucm1)) break;
    }
        
    DIR_RE = DIR_IN;    // Re=Off
    UcmTop = msr_DC_ADC();
    if((ElCm_time & (1<<14)) && ((Ucm0+256) > UcmTop)) fmRx = 1;
    if(ElCm_time < 2) goto discharge_Cm_5mS; 
    
    // ElCm connected    
    fmElCx = 1;
    fsDig3 = 1;     // str[2] = '='
    value_exponent = 6;
    value_units = 'F';
    Utmp = UcmTop;
        
    while(1) {
        delay_ms(100);
        Ux = msr_DC_ADC();    //ADC.2
        Utmp -= Ux;
        Utmp >>= 4;
        Utmp &= 0xFFF;
        dUc_200mS = 0;
        if((Utmp & (1<<11))) goto next;
        
        if(Utmp > 99) Utmp = 99;    // 158.4[mV]
            
        if(fmElCx) dUc_100mS = (char)Utmp;  // dT==100mS
        else {                              // dT==200mS
            dUc_200mS = (char)Utmp;
            if(dUc_200mS <= 1) dUc_200mS = 0;
            dUc_200mS <<= 1;
        }
next:
        fmElCx ^= 1;
        Utmp = Ux;
        if(fmElCx) break;            
    }
        
    // show el_C_sz
    if(dUc_100mS < dUc_200mS) EE_str_off = a_ElCsz;
        
    dUc_200mS = '0';
    while(dUc_100mS > 9) { dUc_100mS -= 10; dUc_200mS++; }
    dUc_100mS += '0';
        
    if(dUc_200mS == '0') { dUc_200mS = dUc_100mS; dUc_100mS = ' '; }
        
    dUcm = UcmTop - Ucm0;
    
    // elC = [(50127-Ucm0)*23400/dUcm-12500]/Re*cnt-49
    // elC = (Ue/Re)/dUc*Tc

#define DCADC_TIME 24872UL  // ADC confersion time    
    lcd_data = (((((50127UL - Ucm0) * DCADC_TIME)
                / dUcm) - 12500UL) / Re0)
                * ElCm_time - 49UL;

    
    if(lcd_data < (3<<8)) {   // elC < 768[nF]
        lcd_data *= 10;
        value_exponent--;
    }
        
discharge_Cm_5mS:        
    DIR_RE = DIR_OUT; RE = 0;   // Re=GND
    delay_ms(5);
    Utmp = msr_DC_ADC();
    DIR_RE = DIR_IN;            // Re=Off
        
    if(!(ElCm_time & (1<<14))) return;
    if(Ucm0 >> 8) return;       // Ucm0 > 256
    if(!(dUcm >> 8)) return;    // dUcm <= 256
    if(Utmp >> 8) return;       // disUcm > 256
        
    if(fmFreqStab) { fmFreqStab = 0; fmRdyOk = 0; }
    else { fmRx = 0; fmIc = 1; }    // mode=Ic
}
/*
void calc_Cx(void)
{
//    lcd_data = (float)Cx * (F1pwr2/F0pwr2 - 1.0);
    kF = F1pwr2 / F0pwr2;
//    lcd_data = (float)Cx * (kF - 1.0) + kF * CK - CK;
    lcd_data = (float)Cx * (kF - 1.0);
    
    EE_str_off = a_Cx;
    value_exponent = 2;
    value_units = 'F';
}
*/
/*
void msrCx_calcLx(void)
{
    float k;
    
    value_units = 'H';
    value_exponent = 7;
    
    if(fmFreqStab) {
        if(fmRdyOk) fmFreqStab = 0;
        msr_Cp();   // res=Cx; set fmLCx!!!
    }
    
    if(!fmLCx) {                // fmLCx==0 
        if(fmC33n) Cx = Cp0;    // C33n==On
        else Cx = C0;           // C33n==Off
    }
    // calculate Lx=(1/(2*pi)^2)/(Cx*F0^2)

    k = 2.533029591e+14F;       // C33n==On
    if(!fmC33n) k *= 10.0;      // C33n==Off
    
    lcd_data = (ulong)(k / ((float)Cx * F0pwr2));
    
    if(fmLCx) EE_str_off = a_lcLx;  // fmLCx==1
    else EE_str_off = a_Lx;

    if(fmC33n) {            // C33n==On
        value_exponent++;
        fsDig3 = 1;         // str[2] = '='
    } else {                // C33n==Off
        delay_ms(500);      // ???????????
        lcd_data -= L0;     // = Lx-L0
    }
}
*/
void msr_Rx(void)
{
    float Ro, dUesr, tUesr0;
    ulong tmp;
    char mask;
    
    fsMinus = 0;
    fsDig3 = 1;     // str[2] = '='
    
    tUesr0 = Uesr0;
    
    if(Kesr0 > Kesr) tUesr0 *= (float)Kesr0;
    else {
        tUesr0 *= (float)Kesr;
        Kesr = Kesr0;
    }
    
    tUesr0 /= (float)Kesr;
    
    msr_Uesr();         // res=Uesr
    
    if(!fmRdyOk) return;

    Ro = (float)L0;
    Ro *= Ro;
    Ro *= (float)Kesr;
    Ro /= (float)KUesr;
    Ro /= 100.0;
    Ro /= tUesr0;

    dUesr = tUesr0 - (float)Uesr;
    if(Uesr > Uesr0) {
        fsMinus = 1;
        dUesr = (float)Uesr - tUesr0;
    }

    Ro *= dUesr;

    if(fmIc) {
        tmp = ((ulong)Ro >> 16) / 3;
        if(Uesr < tmp) fmRdyOk = 0;
        Uesr -= tmp;
        if(Uesr < 29) fmRdyOk = 0;   
    }
    
    lcd_data = (ulong)(Ro/(float)Uesr);

    if(fsMinus) { lcd_data = 0 - lcd_data; return; }
    if(fmRx || fmIc) return;

    // tgCap ElCx ESR
    ElCm_time |= 1;
    tmp = lcd_data * ElCm_time;  // =Resr*ElCm
    tmp >>= 16;
    
    lcd_wrByte(0x48,LCD_CMD);     // Sets_CGRAM_address 0x8
    
    tgCap = 0;
    if(tmp & 0xFF00) tgCap = 0x1F;
        
    tmp &= 0xFF;
    mask = 0xFF;
    while(mask) {
        mask >>= 1;
        if(mask < (char)tmp) {
            tgCap <<= 1;
            tgCap |= 1;
        }
        tgCap &= 0x1F;
        lcd_wrByte(tgCap,LCD_DATA);
    }
}

void main(void)
{
// Port B initialization
DDRB=0x3F;      // LCD interface
// Port C initialization
DDRC=0x00;      // PORTC.5==OUT(CAL)
// Port D initialization
PORTD=0x02;     // RMODE=CMODE
DDRD=0x02;      // PORTD.1==OUT(RMODE)
// Timer/Counter 0 initialization
TCCR0=0x06;     // Clock source: T0 pin Falling Edge
//TCNT0=0x00;
// Timer/Counter 1 initialization
TCCR1A=0x00;    // Normal operation
TCCR1B=0x01;    // clkI/O (No prescaling)
// Timer/Counter 2 initialization
TCCR2=(1<<CS22)|(1<<CS21)|(1<<CS20);    // Normal operation, clkT2S/1024=26mS
// Disable Analog Comparator
ACSR=0x80;
// Disable Pull-up
SFIOR |= (1<<PUD);
// ADC initialization
ADMUX = (1<<REFS0);     // Vref = AVCC with external capacitor at AREF pin
ADCSRA = (1<<ADPS2);    // Division Factor=16
ADCSRA |= (1<<ADEN);    // ADC Enable 

MCUCR |= (1<<ISC01);    // The falling edge of INT0 generates an interrupt request.

    init_lcm();
    ena_btn_int();
    
    while(1)
    {
        if(btn_Mod) RMODE ^= 1;
        btn_Mod = 0;

        if(btn_Set) set_mode();
        
        msr_Freq();     // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (C33n==On -> Fx^2-0.6)
        F0pwr2 = FXpwr2;

        if(dFx > Fx) dFx -= Fx;
        else dFx = Fx - dFx;
        
        if(dFx > 152)   // dFx > 15.2kHz
            fsFreqChange = 1;
        else {          // dFx <= 15.2kHz
            if(fsFreqChange) fmFreqStab = 1;
            fsFreqChange = 0;
        }
        
        dFx = Fx;

        if(Fx > 1050) {  // Fx > 105kHz
            fmRx = 0;
            fmIc = 0;
            fmElCx = 0;
        }
        
        if(btn_Nul) nul_mode();
        
        if(RMODE == CMODE) {
            if(fmC33n) {            // C33n==On
                DIR_RE = DIR_IN;    // Re=Off
                fmC33n = 0;
            }
            fmLCx = 0;
            msr_ElCm();
//            if(!fmElCx) calc_Cx();  // lcd_data = Cx * (F1pwr2/F0pwr2 - 1.0)
            if(!fmElCx) {
                k = F1pwr2 / F0pwr2;
                lcd_data = (float)Cx * (k - 1.0);
                EE_str_off = a_Cx;
                value_exponent = 2;
                value_units = 'F';
            }
//        } else msrCx_calcLx();      // lcd_data = L
        } else {
            value_units = 'H';
            value_exponent = 7;
    
            if(fmFreqStab) {
                if(fmRdyOk) fmFreqStab = 0;
                msr_Cp();   // res=Cx; set fmLCx!!!
            }
    
            if(!fmLCx) {                // fmLCx==0 
                if(fmC33n) Cx = Cp0;    // C33n==On
                else Cx = C0;           // C33n==Off
            }
            // calculate Lx=(1/(2*pi)^2)/(Cx*F0^2) */
            k = 2.533029591e+14F;       // C33n==On
            if(!fmC33n) k *= 10.0;      // C33n==Off
            lcd_data = (ulong)(k / ((float)Cx * F0pwr2));
            if(fmLCx) EE_str_off = a_lcLx;  // fmLCx==1
            else EE_str_off = a_Lx;

            if(fmC33n) {            // C33n==On
                value_exponent++;
                fsDig3 = 1;         // str[2] = '='
            } else {                // C33n==Off
                delay_ms(500);      // ???????????
                lcd_data -= L0;     // = Lx-L0
            }
        }
        
        lcd_LineN = 0;
        
        if(!fmRx && !fmIc) lcd_show_value();  // show "C" or "L" value
        
        lcd_LineN = 1;

        if(fmElCx) {
            msr_Rx();
            value_units = OMEGA;    // "Ohm"
            value_exponent = 11;
            EE_str_off = a_rESR;
            
            if(fmRx) {              // show "Rx"
                EE_str_off = a_rRx;
                lcd_LineN = 0;
            }
            
            if(fmIc) {                  // show "r-ESR"
                EE_str_off = a_icEsr;
                lcd_LineN = 0;
            }
            
            lcd_show_value();
            
            if(fmRx) {                  // show "Uesr"
                lcd_LineN = 1;
                value_units = 'V';
                value_exponent = 11;
                EE_str_off = a_Uesr;
                lcd_data = Uesr;
                lcd_show_value();
            }
            goto is_lcd_LineN;
        }
        
        if(fmLCx) {                     // show "Cp"
            value_units = 'F';
            value_exponent = 2;
            EE_str_off = a_Cp;
                 
            if(fmC33n) {                // C33n==On
                fsDig3 = 1;             //str[2] = '='
                lcd_data = Cx - Cp0;
                if(lcd_data < 63) fmLCx = 0;
            } else lcd_data = Cx - C0;  // C33n==Off
            
            if(lcd_data & (1<<31)) fmLCx = 0;
            if(fmLCx) lcd_show_value();
            goto is_lcd_LineN;
        }

        lcd_LineN = 0;

is_lcd_LineN:
        if(!lcd_LineN) {
            lcd_LineN = 1;
            lcd_EEdata_to_DDRAM(a_0xFF);
        }        
    };
}
